import logging
import re
import socket
import threading
import time

from cmdserver.zeroconf import Zeroconf, ServiceBrowser


logger = logging.getLogger('tivocontroller')


# Mapping of commands to remote codes
CODES = {
    # teleport
    'Home': 'TIVO',
    'My Shows': 'NOWSHOWING',
    'LiveTV': 'LIVETV',
    'Info': 'INFO',
    'Guide': 'GUIDE',
    'Back': 'BACK',
    'Thumbs Up': 'THUMBSUP',
    'Thumbs Down': 'THUMBSDOWN',
    # navigation
    'Up': 'UP',
    'Down': 'DOWN',
    'Left': 'LEFT',
    'Right': 'RIGHT',
    'Return': 'SELECT',
    'Ch+': 'CHANNELUP',
    'Ch-': 'CHANNELDOWN',
    # Play, Pause, FF, Rew
    'Stop': 'STOP',
    'Play': 'PLAY',
    'Rewind': 'REVERSE',
    'Pause': 'PAUSE',
    'FF': 'FORWARD',
    # Skip Back, Slow Motion, Skip Forward
    'Replay': 'REPLAY',
    'Slow': 'SLOW',
    'Advance': 'ADVANCE',
    # Record
    'Record': 'RECORD',
    # colour keys
    'Red': 'ACTION_A',
    'Green': 'ACTION_B',
    'Yellow': 'ACTION_C',
    'Blue': 'ACTION_D',
    # channel selection
    'Clear': 'CLEAR',
    'Enter': 'ENTER',
    # subtitles
    'Subtitles Off': 'CC_OFF',
    'Subtitles On': 'CC_ON'
}

# Named symbols for direct text input -- these work with IRCODE and
# KEYBOARD commands
SYMBOLS = {'-': 'MINUS', '=': 'EQUALS', '[': 'LBRACKET',
           ']': 'RBRACKET', '\\': 'BACKSLASH', ';': 'SEMICOLON',
           "'": 'QUOTE', ',': 'COMMA', '.': 'PERIOD', '/': 'SLASH',
           '`': 'BACKQUOTE', ' ': 'SPACE', '1': 'NUM1', '2': 'NUM2',
           '3': 'NUM3', '4': 'NUM4', '5': 'NUM5', '6': 'NUM6',
           '7': 'NUM7', '8': 'NUM8', '9': 'NUM9', '0': 'NUM0'}

# When in shift mode (with KEYBOARD command only), the same names
# map to a different set of symbols
SHIFT_SYMS = {'_': 'MINUS', '+': 'EQUALS', '{': 'LBRACKET',
              '}': 'RBRACKET', '|': 'BACKSLASH', ':': 'SEMICOLON',
              '"': 'QUOTE', '<': 'COMMA', '>': 'PERIOD', '?': 'SLASH',
              '~': 'BACKQUOTE', '!': 'NUM1', '@': 'NUM2', '#': 'NUM3',
              '$': 'NUM4', '%': 'NUM5', '^': 'NUM6', '&': 'NUM7',
              '*': 'NUM8', '(': 'NUM9', ')': 'NUM0'}


class Tivo(object):
    def __init__(self, tivo):
        self._sock = None
        self.__tivo = tivo
        self.__messages = [''] * 5
        self._messageIdx = 0
        self.currentChannel = ''
        if ':' in tivo['address']:
            address, port = tivo['address'].split(':')
            try:
                port = int(port)
                self.__tivo['address'] = address
                self.__tivo['port'] = port
            except:
                pass
        self.__connect()

    @property
    def name(self):
        return self.__tivo['name']

    @property
    def port(self):
        return self.__tivo['port']

    @property
    def address(self):
        return self.__tivo['address']

    @property
    def version(self):
        return self.__tivo['version']

    def get_tivo_properties(self):
        return {**self.__tivo}

    @property
    def connected(self):
        return self._sock is not None

    def __connect(self):
        """ Connect to the TiVo within five seconds or report error. """
        self.__messages = [''] * 5
        self._messageIdx = 0
        try:
            self._sock = socket.socket()
            self._sock.settimeout(5)
            self._sock.connect((self.address, self.port))
            self._sock.settimeout(None)
            self._statusThread = threading.Thread(name='TivoStatusReader', target=self.__read_from_socket, daemon=True)
            self._statusThread.start()
        except Exception as e:
            self.__record_message('Unable to connect - ' + str(e))
            self.disconnect()

    def disconnect(self):
        if self._sock:
            self.__record_message('Stopping on disconnect')
            self._sock.close()
            self.__record_message('Disconnected')
            self._sock = None
        else:
            logger.info(f"Ignoring disconnect {self.name} has no socket")

    def __send(self, message):
        """ The core output function. Re-connect if necessary, send message, sleep, and check for errors. """
        if not self._sock:
            self.__connect()
        try:
            self._sock.sendall(message.encode())
            time.sleep(0.1)
        except Exception as e:
            logger.error(e)
            self.__record_message('Failed to send ' + str(message) + ' - ' + str(e))
            self.disconnect()
            raise e

    def send_ir(self, sent, *codes):
        """ Expand a command sequence for send(). """
        for each in codes:
            if type(each) == list:
                self.send_ir(sent, *each)
            else:
                if each in CODES:
                    actual = CODES[each]
                    if type(actual) == list:
                        self.send_ir(sent, *actual)
                    else:
                        toSend = 'IRCODE %s\r' % actual
                        sent.append(toSend)
                        self.__send(toSend)
                else:
                    raise ValueError('Unknown IR Code ' + each)
        return {'channel': self.currentChannel, 'sent': sent}

    def set_channel(self, channel):
        """ Sends a single SETCH. """
        toSend = 'SETCH %s\r' % channel
        self.__send(toSend)
        return {'channel': self.currentChannel, 'sent': [toSend]}

    def send_text(self, text):
        """ Expand a KEYBOARD command sequence for send(). """
        sent = []
        for ch in text:
            if 'A' <= ch <= 'Z':
                self.__send_keyboard(sent, 'LSHIFT', ch)
            elif 'a' <= ch <= 'z':
                self.__send_keyboard(sent, ch.upper())
            elif ch in SYMBOLS:
                self.__send_keyboard(sent, SYMBOLS[ch])
            elif ch in SHIFT_SYMS:
                self.__send_keyboard(sent, 'LSHIFT', SHIFT_SYMS[ch])
        return {'channel': self.currentChannel, 'sent': sent}

    def __send_keyboard(self, sent, *codes):
        for each in codes:
            toSend = 'KEYBOARD %s\r' % each
            sent.append(toSend)
            self.__send(toSend)

    def __record_message(self, msg):
        idx = self._messageIdx % 5
        self._messageIdx += 1
        self.__messages[idx] = msg

    def __read_from_socket(self):
        """ Read incoming messages from the socket in a separate thread.
        """
        logger.info(f"[{self.name}] Initialising status reader")
        while self._sock is not None:
            logger.info('Reading')
            try:
                status = self._sock.recv(80).decode()
            except Exception as e:
                status = str(e)
            status = status.strip().title()
            if not status:
                self.__record_message('No status read from socket')
                self.disconnect()
            else:
                self.currentChannel = status
        logger.info(f"[{self.name}] Exiting status reader")

    @property
    def messages(self):
        return self.__messages


class TivoController(object):

    def __init__(self, enabled: bool):
        if enabled:
            self.__tivos = [Tivo(t) for t in self.__find_tivos()]
            self.__ping()
        else:
            self.__tivos = []

    def __ping(self, delay=15):
        time.sleep(delay)
        worker = threading.Thread(name='TivoPinger', target=self.__ping0, daemon=True)
        worker.start()

    def __ping0(self):
        try:
            tivos = self.__find_tivos()
            logger.debug(f"Found {len(tivos)} tivos")
            for new_tivo in tivos:
                existing_tivo = next((t for t in self.__tivos if t.name == new_tivo['name']), None)
                if existing_tivo is None:
                    logger.info(f"Connecting new Tivo {new_tivo['name']} @ {new_tivo['address']}:{new_tivo['port']}")
                    self.__tivos.append(Tivo(new_tivo))
            for old_tivo in self.__tivos:
                dead_tivo = next((t for t in tivos if t['name'] == old_tivo.name), None)
                if dead_tivo is None:
                    logger.info(f"Disconnecting old Tivo {old_tivo.name} @ {old_tivo.address}:{old_tivo.port}")
                    old_tivo.disconnect()
        except:
            logger.exception('Unexpected failure during ping')
        finally:
            self.__ping()

    def has_tivo(self, tivo_name):
        return True if next((t for t in self.__tivos if t.name == tivo_name), None) else False

    def get_tivo(self, tivo_name):
        return next((t for t in self.__tivos if t.name == tivo_name), None)

    def send(self, tivo_name, type, command):
        tivo = self.get_tivo(tivo_name)
        if tivo is not None:
            if type == 'keyboard':
                return tivo.send_text(command)
            elif type == 'ir':
                return tivo.send_ir([], command)
            elif type == 'setch':
                return tivo.set_channel(command)
            else:
                raise ValueError('Unknown command type ' + type)
        else:
            raise ValueError('Unknown tivo ' + tivo_name)

    @property
    def tivos(self):
        return [
            {
                **t.get_tivo_properties(),
                **{'connected': t.connected, 'messages': t.messages, 'channel': t.currentChannel}
            }
            for t in self.__tivos
        ]

    def __find_tivos(self):
        class ZCListener:
            def __init__(self, names):
                self.names = names

            def remove_service(self, server, type, name):
                self.names.remove(name)

            def add_service(self, server, type, name):
                self.names.append(name)

        REMOTE = '_tivo-remote._tcp.local.'

        tivos = []
        tivos_names = []

        # Get the names of TiVos offering network remote control
        serv = Zeroconf()
        try:
            browser = ServiceBrowser(serv, REMOTE, ZCListener(tivos_names))
            # Give them a second to respond
            time.sleep(1)

            # For proxied TiVos, remove the original
            # for t in tivos_names[:]:
            #     if t.startswith('Proxy('):
            #         try:
            #             t = t.replace('.' + REMOTE, '')[6:-1] + '.' + REMOTE
            #             tivos_names.remove(t)
            #         except:
            #             pass

            # Now get the addresses -- this is the slow part
            swversion = re.compile('(\d*.\d*)').findall
            for t in tivos_names:
                s = serv.get_service_info(REMOTE, t)
                if s:
                    name = t.replace('.' + REMOTE, '')
                    address = socket.inet_ntoa(s.address)
                    try:
                        versionProperty = s.properties.get(str.encode('swversion'))
                        if (versionProperty):
                            version = float(swversion(versionProperty.decode())[0])
                        else:
                            version = 0.0
                    except Exception as e:
                        print(e)
                        version = 0.0
                    tivos.append({
                        'name': name,
                        'port': s.port,
                        'version': version,
                        'address': address
                    })

            # For proxies with numeric names, remove the original
            # for name, tivo in tivos.items():
            #     if name.startswith('Proxy('):
            #         address = name.replace('.' + REMOTE, '')[6:-1]
            #         if address in tivos_rev:
            #             tivos.pop(tivos_rev[address])
            return tivos
        except Exception as e:
            print(e)
            return tivos
        finally:
            serv.close()



import logging

from flask_restx import Resource, Namespace

from cmdserver.infoprovider import InfoProvider

logger = logging.getLogger('playingnow')

api = Namespace('1/playingnow', description='Gets info about what is playing now')


@api.route('')
class PlayingNow(Resource):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__info_provider: InfoProvider = kwargs['info_provider']

    def get(self):
        return self.__info_provider.info, 200



import logging

from flask_restx import Resource, Namespace

from cmdserver.infoprovider import InfoProvider

logger = logging.getLogger('wake')

api = Namespace('1/wake', description='Sends a WOL to the target MC server')


@api.route('')
class Wake(Resource):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__info_provider: InfoProvider = kwargs['info_provider']

    def put(self):
        try:
            if self.__info_provider.wake():
                return '', 200
            else:
                return 'No MAC address', 400
        except Exception as e:
            return str(e), 500



import faulthandler
import os
from os import path
from typing import Tuple

from autobahn.twisted.resource import WebSocketResource
from flask import Flask
from flask_restx import Api

from cmdserver.pjcontroller import PJController
from cmdserver.apis import command, commands, playingnow, tivo, tivos, pj, info, version, wake
from cmdserver.commandcontroller import CommandController
from cmdserver.config import Config
from cmdserver.infoprovider import InfoProvider
from cmdserver.tivocontroller import TivoController
from cmdserver.ws import WsServer

API_PREFIX = '/api/1'

faulthandler.enable()
if hasattr(faulthandler, 'register'):
    import signal

    faulthandler.register(signal.SIGUSR2, all_threads=True)


def create_app(cfg: Config) -> Tuple[Flask, 'WsServer']:
    ws_server = WsServer()
    info_provider = InfoProvider(cfg, ws_server)
    resource_args = {
        'command_controller': CommandController(cfg),
        'tivoController': TivoController(cfg.use_tivo),
        'pj_controller': PJController(cfg),
        'info_provider': info_provider,
        'config': cfg,
        'version': cfg.version
    }
    app = Flask('cmdserver')
    api = Api(app, prefix='/api', doc='/api/doc/', version=resource_args['version'], title='cmdserver',
              description='Backend api for cmdserver')

    def decorate_ns(ns, p=None):
        for r in ns.resources:
            r.kwargs['resource_class_kwargs'] = resource_args
        api.add_namespace(ns, path=p)

    decorate_ns(commands.api)
    decorate_ns(command.api)
    decorate_ns(playingnow.api)
    decorate_ns(tivos.api)
    decorate_ns(tivo.api)
    decorate_ns(info.api)
    decorate_ns(pj.api)
    decorate_ns(version.api)
    decorate_ns(wake.api)
    return app, ws_server


def main(args=None):
    """ The main routine. """
    cfg = Config('cmdserver')
    app, ws_server = create_app(cfg)
    logger = cfg.configure_logger()

    if cfg.useTwisted:
        import logging
        logger = logging.getLogger('twisted')
        from twisted.internet import reactor
        from twisted.web.resource import Resource
        from twisted.web import static, server
        from twisted.web.wsgi import WSGIResource
        from twisted.application import service
        from twisted.internet import endpoints

        class ReactApp:
            """
            Handles the react app (excluding the static dir).
            """

            def __init__(self, path):
                # TODO allow this to load when in debug mode even if the files don't exist
                self.publicFiles = {f: static.File(os.path.join(path, f)) for f in os.listdir(path) if
                                    os.path.exists(os.path.join(path, f))}
                self.indexHtml = ReactIndex(os.path.join(path, 'index.html'))

            def get_file(self, path):
                """
                overrides getChild so it always just serves index.html unless the file does actually exist (i.e. is an
                icon or something like that)
                """
                return self.publicFiles.get(path.decode('utf-8'), self.indexHtml)

        class ReactIndex(static.File):
            """
            a twisted File which overrides getChild so it always just serves index.html (NB: this is a bit of a hack,
            there is probably a more correct way to do this but...)
            """

            def getChild(self, path, request):
                return self

        class FlaskAppWrapper(Resource):
            """
            wraps the flask app as a WSGI resource while allow the react index.html (and its associated static content)
            to be served as the default page.
            """

            def __init__(self):
                super().__init__()
                self.wsgi = WSGIResource(reactor, reactor.getThreadPool(), app)
                import sys
                if getattr(sys, 'frozen', False):
                    # pyinstaller lets you copy files to arbitrary locations under the _MEIPASS root dir
                    uiRoot = os.path.join(sys._MEIPASS, 'ui')
                elif cfg.webappPath is not None:
                    uiRoot = cfg.webappPath
                else:
                    # release script moves the ui under the analyser package because setuptools doesn't seem to include
                    # files from outside the package
                    uiRoot = os.path.join(os.path.dirname(__file__), 'ui')
                logger.info('Serving ui from ' + str(uiRoot))
                self.react = ReactApp(uiRoot)
                self.static = static.File(os.path.join(uiRoot, 'static'))
                self.icons = static.File(cfg.iconPath)
                ws_server.factory.startFactory()
                self.ws_resource = WebSocketResource(ws_server.factory)

            def getChild(self, path, request):
                """
                Overrides getChild to allow the request to be routed to the wsgi app (i.e. flask for the rest api
                calls), the static dir (i.e. for the packaged css/js etc), the various concrete files (i.e. the public
                dir from react-app), the command icons or to index.html (i.e. the react app) for everything else.
                :param path:
                :param request:
                :return:
                """
                # allow CORS (CROSS-ORIGIN RESOURCE SHARING) for debug purposes
                request.setHeader('Access-Control-Allow-Origin', '*')
                request.setHeader('Access-Control-Allow-Methods', 'GET, PUT')
                request.setHeader('Access-Control-Allow-Headers', 'x-prototype-version,x-requested-with')
                request.setHeader('Access-Control-Max-Age', '2520')  # 42 hours
                logger.debug(f"Handling {path}")
                if path == b'ws':
                    return self.ws_resource
                elif path == b'api':
                    request.prepath.pop()
                    request.postpath.insert(0, path)
                    return self.wsgi
                elif path == b'static':
                    return self.static
                elif path == b'icons':
                    return self.icons
                else:
                    return self.react.get_file(path)

            def render(self, request):
                return self.wsgi.render(request)

        application = service.Application('cmdserver')
        if cfg.is_access_logging is True:
            site = server.Site(FlaskAppWrapper(), logPath=path.join(cfg.config_path, 'access.log').encode())
        else:
            site = server.Site(FlaskAppWrapper())
        endpoint = endpoints.TCP4ServerEndpoint(reactor, cfg.port, interface='0.0.0.0')
        endpoint.listen(site)
        reactor.run()
    else:
        logger.error('Icons are not available in debug mode')
        # get config from a flask standard place not our config yml
        app.run(debug=cfg.run_in_debug, host='0.0.0.0', port=cfg.port, use_reloader=False)


if __name__ == '__main__':
    main()



import logging
from typing import Callable, Optional, List

from autobahn.exception import Disconnected
from autobahn.twisted import WebSocketServerProtocol, WebSocketServerFactory

logger = logging.getLogger('ws')


class WsServer:

    def __init__(self):
        self.__factory = WsServerFactory()

    @property
    def factory(self) -> 'WsServerFactory':
        return self.__factory

    def broadcast(self, msg: str):
        self.__factory.broadcast(msg)


class WsProtocol(WebSocketServerProtocol):

    def onConnect(self, request):
        logger.info(f"Client connecting: {request.peer}")

    def onOpen(self):
        logger.info("WebSocket connection open")
        self.factory.register(self)

    def onClose(self, was_clean, code, reason):
        logger.info(f"WebSocket connection closed: clean? {was_clean}, code: {code}, reason: {reason}")

    def onMessage(self, payload, is_binary):
        try:
            s = payload.decode('utf-8')
            logger.info(f"Received {s}")
        except:
            logger.exception('Message received failure')


class WsServerFactory(WebSocketServerFactory):
    protocol = WsProtocol

    def __init__(self, *args, **kwargs):
        super(WsServerFactory, self).__init__(*args, **kwargs)
        self.__clients: List[WsProtocol] = []
        self.__state_provider: Optional[Callable[[], str]] = None

    def init(self, state_provider: Callable[[], str]):
        self.__state_provider = state_provider

    def register(self, client: WsProtocol):
        if client not in self.__clients:
            logger.info(f"Registered client {client.peer}")
            self.__clients.append(client)
            if self.__state_provider:
                state = self.__state_provider()
                if state:
                    client.sendMessage(state.encode('utf8'), isBinary=False)
        else:
            logger.info(f"Ignoring duplicate client {client.peer}")

    def unregister(self, client: WsProtocol):
        if client in self.__clients:
            logger.info(f"Unregistering client {client.peer}")
            self.__clients.remove(client)
        else:
            logger.info(f"Ignoring unregistered client {client.peer}")
    
    def broadcast(self, msg: str):
        logger.debug(f"Broadcasting {msg}")
        if self.__clients:
            disconnected_clients = []
            for c in self.__clients:
                logger.debug(f"Sending to {c.peer}")
                try:
                    c.sendMessage(msg.encode('utf8'), isBinary=False)
                except Disconnected as e:
                    logger.exception(f"Failed to send to {c.peer}, discarding")
                    disconnected_clients.append(c)
            for c in disconnected_clients:
                self.unregister(c)
        else:
            logger.debug(f"No devices connected, ignoring {msg}")



import logging

from flask_restx import Resource, Namespace

logger = logging.getLogger('command')

api = Namespace('1/command', description='Provides ability to execute a configured command')


@api.route('/<string:command>')
class Command(Resource):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__controller = kwargs['command_controller']

    def put(self, command):
        logger.info(f'Executing {command}')
        result = self.__controller.execute(command)
        if result is None:
            logger.info(f'Unknown {command}')
            return None, 404
        else:
            if result[0] == 0:
                logger.info(f'Executed {command} successfully')
                return None, 200
            else:
                logger.info(f'Executed {command} with unexpected result {result[0]}')
                return {'errorCode': result[0]}, 500



import logging

import requests
from plumbum import local


logger = logging.getLogger('commandcontroller')


class CommandController:

    def __init__(self, config):
        actual_commands = {commandId: command for commandId, command in config.commands.items() if commandId != 'defaults'}
        defaults = config.commands['defaults'] if 'defaults' in config.commands else None
        self.__commands = {command_id: self.__add_defaults(command_id, command, defaults) for command_id, command in actual_commands.items()}
        self.__launchers = {command_id: self.__get_launcher(command_id, command, defaults) for command_id, command in actual_commands.items()}

    @staticmethod
    def __add_defaults(command_id, command, defaults):
        if 'icon' not in command:
            command['icon'] = command_id + '.ico'
        if 'zoneId' not in command and 'zoneId' in defaults:
            command['zoneId'] = defaults['zoneId']
        if 'volume' not in command and 'volume' in defaults:
            command['volume'] = defaults['volume']
        if 'stopAll' not in command and 'stopAll' in defaults:
            command['stopAll'] = defaults['stopAll']
        return command

    @staticmethod
    def __get_launcher(command_id, command, defaults):
        if 'remote' in command:
            return RemoteCommandExecutor(f"{defaults['remote_prefix']}/{command_id}")
        else:
            exe = defaults['exe'] if defaults is not None and 'exe' in defaults else command['exe']
            return local[exe][command['args']] if 'args' in command else local[exe]

    @property
    def commands(self):
        return self.__commands

    def get_command(self, command_id):
        return self.__commands[command_id] if command_id in self.__commands else None

    def execute(self, command_id):
        command = self.get_command(command_id)
        if command is not None:
            result = self.__launchers[command_id]['executor'].run(retcode=None)
            logger.info(f"Executed command {command_id}, result is {result[0]}")
            logger.info('Command output: ')
            logger.info(result[1])
            return result
        return None


class RemoteCommandExecutor:

    def __init__(self, address: str):
        self.__address = address
        logger.info(f"Created RemoteCommandExecutor for {self.__address}")

    def __getitem__(self, item):
        # hack to allow the execute command to work as expected
        return self

    def run(self, **kwargs):
        r = requests.put(self.__address)
        if r.status_code == 200:
            return [0, '']
        else:
            return [2, f"{r.status_code} - {r.text}"]



import logging

from flask_restx import Resource, Namespace

from cmdserver.jvccommands import get_all_command_info
from cmdserver.pjcontroller import PJController

logger = logging.getLogger('info')

api = Namespace('1/info', description='Gets info about commands supported by the pj (if any)')


@api.route('')
class Info(Resource):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__pj_controller: PJController = kwargs['pj_controller']
        self.__supported = get_all_command_info() if self.__pj_controller.enabled else []

    def get(self):
        if self.__supported is None:
            return None, 500
        else:
            return self.__supported, 200



import logging
from enum import Enum
from threading import Lock
from time import sleep

from cmdserver.debounce import debounce
from cmdserver.jvc import CommandExecutor, CommandNack
from cmdserver.jvccommands import Command, load_all_commands, Numeric

logger = logging.getLogger('pjcontroller')


class PJController:

    def __init__(self, config):
        self.__pj_macros = config.pj_macros
        self.__executor = CommandExecutor(host=config.pj_ip) if config.pj_ip else None
        self.__commands = load_all_commands()
        self.__lock = Lock()

    @property
    def enabled(self):
        return self.__executor is not None

    def __connect(self):
        self.__executor.connect()

    @debounce(4)
    def __disconnect(self):
        self.__executor.disconnect(fail=False)

    def get(self, command):
        with self.__lock:
            try:
                cmd = Command[command]
                self.__connect()
                val = self.__executor.get(cmd)
                self.__disconnect()
                if val is not None:
                    return val.name if isinstance(val, Enum) else val.replace('"', '').strip()
                return val
            except KeyError:
                logger.warning(f"Ignoring unknown command {command}")
                return None
            except CommandNack:
                logger.exception(f"Command NACKed - GET {command}")
                return -1
            except:
                logger.exception(f"Unexpected failure while executing cmd: {command}")
                return -1

    def send(self, commands):
        """ Sends the commands to the PJ """
        with self.__lock:
            vals = []
            self.__connect()
            for command in commands:
                if command[0:5] == 'PAUSE':
                    sleep_secs = float(command[5:])
                    logger.info(f"Sleeping for {sleep_secs:.3f}")
                    sleep(sleep_secs)
                else:
                    if command in self.__pj_macros:
                        for cmd in self.__pj_macros[command]:
                            vals.append(self.__execute(cmd))
                    else:
                        vals.append(self.__execute(command))
            self.__disconnect()
            return vals

    def __execute(self, cmd):
        tokens = cmd.split('.')
        if len(tokens) > 1:
            try:
                cmd_enum = Command[tokens[0]]
                if isinstance(cmd_enum.value, tuple):
                    cmd_arg_enum = cmd_enum.value[1]
                    if cmd_arg_enum.__name__ == tokens[1]:
                        logger.info(f"Executing {cmd}")
                        if issubclass(cmd_arg_enum, Enum):
                            return self.__executor.set(cmd_enum, cmd_arg_enum[tokens[2]])
                        elif issubclass(cmd_arg_enum, Numeric):
                            return self.__executor.set(cmd_enum, Numeric(int(tokens[2])))
                        else:
                            logger.warning(f"Unsupported value type for {cmd} - {cmd_arg_enum.__name__}")
            except (AttributeError, KeyError):
                logger.exception(f"Ignoring unknown command {cmd}")
            except:
                logger.exception(f"Unexpected exception while processing {cmd}")
        else:
            logger.error(f"Ignoring unknown command {cmd}")



import logging

from flask import request
from flask_restx import Resource, Namespace

from cmdserver.pjcontroller import PJController

logger = logging.getLogger('pj')

api = Namespace('1/pj', description='Controls a JVC PJ')


@api.route('/<string:command>')
class PJ(Resource):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__pj_controller: PJController = kwargs['pj_controller']

    def get(self, command):
        if self.__pj_controller.enabled:
            logger.info(f">> GET {command}")
            result = self.__pj_controller.get(command)
            logger.info(f"<< GET {command} = {result}")
            if result is None:
                return None, 404
            elif result == -1:
                return None, 500
            else:
                return result, 200
        else:
            return None, 501


@api.route('')
class UpdatePJ(Resource):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__pj_controller: PJController = kwargs['pj_controller']

    def put(self):
        if self.__pj_controller.enabled:
            payload = request.get_json()
            logger.info(f"Executing {payload}")
            result = self.__pj_controller.send(payload)
            if result is None:
                logger.info(f"Unknown command {payload}")
                return None, 404
            else:
                return None, 200
        else:
            return None, 501



import logging

from flask_restx import Resource, Namespace

logger = logging.getLogger('tivo')

api = Namespace('1/tivo', description='Gets info about a named tivo')


@api.route('/<string:tivo>')
class Tivo(Resource):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__controller = kwargs['tivoController']

    def get(self, tivo):
        tivo = self.__controller.get_tivo(tivo)
        if tivo is None:
            return 404
        else:
            return {'channel': tivo.currentChannel, 'messages': tivo.messages, 'connected': tivo.connected}, 200
